module ArrayBoard (
         module PiecesAndPlayers,
         Board (ArrayBoard),
         initBoard, find, change, findAll
       ) where


import Data.List hiding (find, sort)
import Data.Array
import Sort
import PiecesAndPlayers
import qualified ListBoard



newtype Board
       = ArrayBoard (Array (Int , Int) State)
       deriving (Eq)


arrayToList
  :: Board -> ListBoard.Board
arrayToList (ArrayBoard a)
  = ListBoard.ListBoard (
      reverse (
        transpose (
          map (map snd) (
            groupBy (\ p q -> fst (fst p) == fst (fst q)) (assocs a)
          )
        )
      )
    )

listToArray
  :: ListBoard.Board -> Board
listToArray (ListBoard.ListBoard brd)
  = ArrayBoard (array ((1 , 1) , (8 , 8)) (concat (ListBoard.indexBy (,) brd)))


instance Show Board
  where

    showsPrec n
      = showsPrec n . arrayToList


instance Read Board
  where

    readsPrec n
      = map (\ (brd , t) -> (listToArray brd , t)) . readsPrec n


initBoard
  :: Board
initBoard
  = listToArray ListBoard.initBoard

find
  :: Int -> Int -> Board -> State
find i j (ArrayBoard a)
  = a ! (i , j)

change
  :: State -> Int -> Int -> Board -> Board
change st i j (ArrayBoard a)
  = ArrayBoard (a // [((i , j) , st)])

findAll
  :: Player -> Board -> [(Piece , Sq)]
findAll pl (ArrayBoard a)
  = let
      interesting (_ , j)
        = case j of
            Just (Piece pl' _)
              -> pl == pl'
            _
              -> False
    in
    sort (
      map (\ (q , Just (Piece _ p)) -> (p , uncurry square q)) (
        filter interesting (assocs a)
      )
    )

