-- #!/usr/bin/runhugs -98  -- runhugsiga jooksutamiseks kommenteeri see lahti
                           -- (vajadusel paranda tistee igeks vastavalt
                           -- sellele, kus  runhugs  Teie arvutis paikneb).

module Main where



import System.IO
import qualified FPKodu3  -- Kvalifitseerimine on selleks, et  move  jm 
                          -- mooduli Minimax samanimeliste muutujatega 
                          -- kollisiooni ei tekitaks.
import Minimax


{-
Abidefinitsioonid.
-}


-- Seisu-, kigu- ja partiitbid.

type Position
       = FPKodu3.Position

type Move
       = FPKodu3.Move

type Play
       = [Node Position Move Float]


-- Partii initsialiseerimine (alustades antud positsioonist).

initPlay
  :: Position -> Play
initPlay pos
  = [(pos , initTree pos)]



{-
Jrgnev lhike osa seob meie malemngu moodulis Minimax defineeritud ldiste 
suurustega.
-}


-- Defineerime minimaksile vajalike muutujate vrtused.

instance Game Position Move Float
  where

    move
      = FPKodu3.move

    allMoves
      = FPKodu3.allMoves

    value
      = FPKodu3.value


-- Analsiv funktsioon.

analyze
  :: Position -> AnalyzeTree Move Float -> AnalyzeTree Move Float
analyze
  = minimax 5  -- Argumendiks on poolkikude arv, kui palju ette vaadatakse.
               -- Parameetriga 5 suudab programm lahendada kuni 3-kigulisi 
               -- matilesandeid (kus alustaja matistab).




{-
Siit algab suhtlusega seotud osa.

Kik Xboardiga suhtlemine kib lbi mbersuunatud standardsisendi ja
-vljundi (mbersuunamisega tegeleb Xboard). Seega kik standardsisendist
lugemised thendavad Xboardi saadetud teadete lugemist ja kik
standardvljundisse kirjutamised saadavad teateid Xboardile.

See programm ei suuda teateid vastu vtta kigu mtlemise ajal.
-}


-- Xboardiga suhtlemisel tuleb programmil saata talle oma erisuste nimekiri,
-- mis on jrgmine.

features
  :: [String]
features
  = [
     "setboard=1"
               -- et  xboard  saadaks algseisu FEN-notatsioonis
               -- ksuga  setboard
    ,
     "usermove=1"
               -- et  xboard  saadaks kasutajakigud meie programmile
               -- kujul  usermove <MOVE> , kus <MOVE> on kik
    ,
     "variants=\"normal\""
               -- tunneme ainult tavamalet
    ,
     "sigint=0 sigterm=0"
               -- signaalide ttlus puudub
    ,
     "time=0"
               -- kellaajaga ei oska me midagi peale hakata
    ,
     "draw=0"
               -- ksku  draw  ei tunne
    ,
     "colors=0"
               -- seda ksku me ka ei tunne
    ,
     "analyze=0"
               -- analsiga meid mitte tlitada
    ,
     "myname=\"FPKodu3\""
               -- meie programmi nimi :)
    ,
     "name=0"
               -- vastase nime meile mitte saata
    ,
     "done=1"
               -- rohkem parameetreid pole
    ]


-- Ka mngu lpust tuleb programmil Xboardile teada anda, selleks on kindel 
-- formaat.

loseCode FPKodu3.White
  = "0-1"
loseCode _
  = "1-0"

endCode pl (Definite end _)
  = case end of
      Lose
        -> loseCode pl
      _
        -> "1/2-1/2"



-- Kui muutuja   isDebugMode  vrtus on  True , kantakse infot programmi 
-- tegevuse kohta logifaili. Vajadusel varieerige seda, kumb vrtus on 
-- nhtav. Jrgnev rida muutujaid on seotud logide kirjutamisega.

isDebugMode
  :: Bool
isDebugMode
  = False
    -- True

logi
  :: String -> IO ()
logi
  = appendFile "logi.fp3"

cmdLogi
  :: String -> IO ()
cmdLogi cmd
  = logi ("Received  " ++ cmd ++ " .\n")

startLogi, newLogi
  :: IO ()

startLogi
  = logi "START OF A NEW SESSION\n"

newLogi
  = logi "START OF A NEW GAME\n"

analysisLogi
  :: Position -> AnalyzeTree Move Float -> IO ()
analysisLogi pos (Branch mvs val)
  = logi ("ANALYSIS AT " ++ shows pos "\n") >>
    mapM_ (
      \ (mv , Branch _ val')
        -> logi (FPKodu3.showMove pos mv) >> 
           logi ": " >> 
           logi (shows val' "\n")
    ) mvs

genLogi
  :: IO () -> IO ()
genLogi xLogi
  = if isDebugMode
       then xLogi
       else return ()


-- Jrgnevad meie programmi konkreetsed tegevused.

-- go  mtleb etteantud partii jtkuna programmi kigu, sooritab selle ning 
-- siirdub aktiivsesse reiimi.

go
  :: Play -> IO ()
go play@((pos , at) : _)
  = let
      at'@(Branch mvs res)
        = analyze pos at
      (mv , next@(pos' , Branch mvs' res'))
        = chooseBest (pos , at')
    in
    genLogi (analysisLogi pos at') >>
    if null mvs
       then hPutStrLn stdout (endCode (FPKodu3.getPlayer pos) res) >>
            active play
       else hPutStrLn stdout ("move " ++ FPKodu3.showMove pos mv) >> 
            (if null mvs'
               then hPutStrLn stdout (endCode (FPKodu3.getPlayer pos') res')
               else return ()) >>
            active (next : play)


-- usermove  sooritab ettantud kigu etteantud partii jtkuks ja siirdub 
-- edasi etteanud tegevust tegema.

usermove
  :: String -> Play -> (Play -> IO ()) -> IO ()
usermove arg play@(current@(pos , _) : _) mode
  = genLogi (logi ("usermove arg is  " ++ arg ++ " .\n")) >>
    mode (makeGiven (FPKodu3.readMove pos arg) FPKodu3.showMove current : play)


-- undo  vtab etteantud partiis kigu tagasi ja siirdub passiivsesse
-- reiimi.

undo
  :: Play -> IO ()
undo play@(_ : past)
  = if null past
       then passive play
       else passive past


-- Nd tulevad Xboardilt tulnud kskude analsimise ja sobiva reaktsiooni
-- valimisega seotud osad.

-- Xboardi saadetud ksk vib olla 0 vi 1 argumendiga, viimasel juhul on ksu 
-- ja argumendi vahel thik. Funktsioon  getCmd  parsib Xboardilt saadud rea
-- ksuks ja argumendiks (viimane on "", kui argument puudub).

getCmd
  :: IO (String , String)
getCmd
  = do
      s <- hGetLine stdin
      let p@(cmd , rest@ ~(_ : arg)) = break (== ' ') s
      genLogi (cmdLogi cmd)
      return (cmd , if null rest then "" else arg)


-- other  reageerib universaalsetele kskudele nagu  quit  ja  new , muudel 
-- juhtudel sooritab etteantud tegevuse.

other
  :: String -> IO () -> IO ()
other cmd act
  = case cmd of
      "quit"
        -> return ()
      "new"
        -> genLogi newLogi >>
           active (initPlay FPKodu3.initPosition)
      _
        -> act


-- Aktiivne reiim, kus Xboardilt tulevad ainult programmi vastase
-- (kasutaja, teine programm) kigud. Passiivsesse reiimi viib ksk  force .

active
  :: Play -> IO ()
active play
  = do
      (cmd , arg) <- getCmd
      case cmd of
        "usermove"
          -> usermove arg play go
        "go"
          -> go play
        "force"
          -> passive play
        _
          -> other cmd (active play)


-- Passiivne reiim, kus kigud tulevad Xboardilt etteantuna. Aktiivsesse
-- reiimi viib ksk  go .

passive
  :: Play -> IO ()
passive play
  = do
      (cmd , arg) <- getCmd
      case cmd of
        "setboard"
          -> passive (initPlay (read arg :: Position))
        "usermove"
          -> usermove arg play passive
        "go"
          -> go play
        "undo"
          -> undo play
        _
          -> other cmd (passive play)


-- Sessiooni alustamine.

start
  :: IO ()
start
  = do
      (cmd , _) <- getCmd
      case cmd of
        "xboard"
          -> hPutStrLn stdout (unlines (map ("feature " ++) features)) >> 
             start
        _
          -> other cmd start


-- Phiprogramm.

main
  :: IO ()
main
  = hSetBuffering stdout NoBuffering >>
    genLogi startLogi >>
    start

