module FunBoard (
         module PiecesAndPlayers,
         Board (),
         initBoard, find, change, findAll
       ) where


import Data.Array
import PiecesAndPlayers
import qualified ListBoard
import qualified ArrayBoard



newtype Board
       = FunBoard (Int -> Int -> State)


funToList
  :: Board -> ListBoard.Board
funToList (FunBoard f)
  = ListBoard.ListBoard (
      map (map (uncurry f)) [ [ (i , j) | i <- [1 .. 8] ] | j <- [8, 7 .. 1] ]
    )

arrayToFun
  :: ArrayBoard.Board -> Board
arrayToFun (ArrayBoard.ArrayBoard a)
  = FunBoard (curry (a !))


instance Eq Board
  where

    f == g
      = funToList f == funToList g


instance Show Board
  where

    showsPrec n
      = showsPrec n . funToList


instance Read Board
  where

    readsPrec n
      = map (\ (a , t) -> (arrayToFun a , t)) . readsPrec n


initBoard
  :: Board
initBoard
  = arrayToFun ArrayBoard.initBoard

find
  :: Int -> Int -> Board -> State
find i j (FunBoard f)
  = f i j

change
  :: State -> Int -> Int -> Board -> Board
change st i j (FunBoard f)
  = let
      f' i' j'
        = if i' == i && j' == j
             then st
             else f i' j'
    in
    FunBoard f'

findAll
  :: Player -> Board -> [(Piece , Sq)]
findAll pl
  = ListBoard.findAll pl . funToList


