module Minimax (
         module Result,
         module Minimax
       ) where


import Prelude hiding (negate)
import Result
import Sort



class (Eq pos, Show pos, Eq mv, RealFrac a)
  => Game pos mv a
  |  pos -> mv
  ,  pos -> a
  where
    move :: mv -> pos -> pos
    allMoves :: pos -> [(mv , pos)]
    value :: pos -> Value a
    

data RealFrac a
  => AnalyzeTree mv a
       = Branch [(mv , AnalyzeTree mv a)] (Value a)


getVal
  :: RealFrac a
  => (mv , AnalyzeTree mv a) -> Value a
getVal (_ , Branch _ v)
  = v

valCmp
  :: RealFrac a
  => (mv , AnalyzeTree mv a) -> (mv , AnalyzeTree mv a) -> Ordering
valCmp p q
  = compare (getVal p) (getVal q)

initTree
  :: Game pos mv a
  => pos -> AnalyzeTree mv a
initTree pos
  = Branch 
      (map (\ (mv , pos) -> (mv , initTree pos)) (allMoves pos))
        (value pos)

lift
  :: RealFrac a
  => Value a -> Value a
lift (Definite res n)
  = Definite (negate res) (n + 1)
lift (Estimated x)
  = Estimated (- x)

minimax
  :: Game pos mv a
  => Int -> pos -> AnalyzeTree mv a -> AnalyzeTree mv a
minimax d pos
  | d < 0
    = error "minimax: negative argument!"
  | otherwise
    = foldr (.) id (map (flip (ab (immediate Win)) pos) [d, d - 1 .. 1])

mapAccumUntilCont acc p f (x : xs) c
  = f x acc $
    \ y acc' ->
      if p y
         then c (y : xs) (if null xs then acc' else Estimated (toRealFrac acc'))
         else mapAccumUntilCont acc' p f xs $
              c . (y :)
mapAccumUntilCont acc _ _ _        c
  = c [] acc

ab
  :: Game pos mv a
  => Value a -> Int -> pos -> AnalyzeTree mv a -> AnalyzeTree mv a
ab upVal d pos at@(Branch mvs val)
  | d > 0
    = case val of
        Estimated _
          | null mvs
            -> at
          | otherwise
            -> let
                 p = (> upVal) . lift . getVal
                 f (mv , au) curVal c
                   = let
                       au'@(Branch _ curVal')
                         = ab curVal (d - 1) (move mv pos) au
                     in
                     c (mv , au') (min curVal curVal')
               in
               mapAccumUntilCont (immediate Win) p f (sortBy valCmp mvs) $
               \ mvs' val'
                 -> Branch mvs' (lift val')
        _
          -> at
  | otherwise
    = at

type Node pos mv a
       = (pos , AnalyzeTree mv a)

makeGiven
  :: Game pos mv a
  => mv -> (pos -> mv -> String) -> Node pos mv a -> Node pos mv a
makeGiven mv showMove (pos , Branch mvs val)
  = let
      (_ , at)
        = let
            mvs' = filter ((== mv) . fst) mvs
          in
          if null mvs'
             then error (
                    "makeGiven: " ++ 
                    showMove pos mv ++ 
                    ": illegal move in position  " ++
                    show pos
                  )
             else head mvs'
    in
    (move mv pos , at)

chooseBest
  :: Game pos mv a
  => Node pos mv a -> (mv , Node pos mv a)
chooseBest (pos , Branch mvs val)
  = let
      (mv , at)
        = head (filter ((== val) . lift . getVal) mvs)
    in
    (mv , (move mv pos , at))

