-- See moodul defineerib malendi- ja mngijatbi, mned nende 
-- kombineerimisel saadavad tbid ning mned operatsioonid nendel.

module PiecesAndPlayers (
         module Square,
         Piece (..),
         allPieces, crownPieces, pieceSyms,
         Player (..),
         op, plDir, relRank,
         PiecesList,
         remove, add,
         PlPiece (..),
         State
       ) where


import Data.Char
import Data.Maybe
import Data.List
import Square



{-
Malendid.
-}

data Piece
       = King    -- Kuningas,
       | Queen   -- lipp,
       | Rook    -- vanker,
       | Bishop  -- oda,
       | Knight  -- ratsu,
       | Pawn    -- ettur.
       deriving (Bounded, Enum, Eq, Ord)


-- List kigist malenditest tbidefinitsioonis esitatud jrjekorras.

allPieces
  :: [Piece]
allPieces
  = [minBound .. maxBound]


-- List kigist malenditest, milleks vib ettur moonduda.

crownPieces
  :: [Piece]
crownPieces
  = [Queen .. Knight]


-- Malendite thised FEN-notatsioonis.

pieceSyms
  :: [Char]
pieceSyms
  = "kqrbnp"


-- Malendile vastava FEN-kuju leidmine.

instance Show Piece
  where

    showsPrec _ p
      = (:) (fromJust (lookup p (zip allPieces pieceSyms)))


-- Malendi FEN-kuju jrgi malendi tuvastamine.

instance Read Piece
  where

    readsPrec _ (s@ ~(c : cs))
      | null s
        = []
      | otherwise
        = case lookup c (zip pieceSyms allPieces) of
            Just p
              -> [(p , cs)]
            _
              -> []


{-
Mngijad.
-}

data Player
       = White      -- Valge,
       | Black      -- must.
       deriving (Eq)


-- Mngijale vastava FEN-kuju leidmine.

instance Show Player
  where

    showsPrec _ White
      = ('w' :)
    showsPrec _ Black
      = ('b' :)


-- Mngijale vastava FEN-kuju jrgi mngija tuvastamine.

instance Read Player
  where

    readsPrec _ (c : cs)
      | c == 'w'
        = [(White , cs)]
      | c == 'b'
        = [(Black , cs)]
      | otherwise
        = []
    readsPrec _ _
      = []


-- Antud mngija vastase leidmine.

op
  :: Player -> Player
op White
  = Black
op _
  = White


-- Antud mngija etturite liikumissuunda nitava tisarvu leidmine.

plDir :: Player -> Int
plDir White
  = 1
plDir _
  = -1


-- Antud mngija poolt vaadates esimese rea jrjekorranumbri leidmine.

relRank
  :: Player -> Int -> Int
relRank White rw
  = rw
relRank _     rw
  = 9 - rw



{-
Mngija malendite ja vljade loetelu.
-}

type PiecesList
        = ([(Piece , Sq)] , [(Piece , Sq)])


-- Mned abifunktsioonid.

deleteInSorted
  :: Ord a
  => a -> [a] -> [a]
deleteInSorted x xs@(z : zs)
  = case compare x z of
      GT
        -> z : deleteInSorted x zs
      EQ
        -> zs
      _
        -> xs

plVariants
  :: Player -> (a -> a) -> (a , a) -> (a , a)
plVariants pl f (wps , bps)
  = case pl of
      White
        -> (f wps , bps)
      _
        -> (wps , f bps)


-- Antud koordinaatidega viidataval vljal asuva antud mngija antud malendi
-- kustutamine malendite ja vljade loetelust.

remove
  :: Piece -> Int -> Int -> Player -> PiecesList -> PiecesList
remove p i j pl
  = plVariants pl (deleteInSorted (p , square i j))


-- Antud koordinaatidega viidatavale vljale antud mngija antud malendi
-- lisamine malendite ja vljade loetellu.

add
  :: Piece -> Int -> Int -> Player -> PiecesList -> PiecesList
add p i j pl
  = plVariants pl (insert (p , square i j))



{-
Mngija malendid.
-}

data PlPiece
       = Piece Player Piece
       deriving (Eq)


-- Mngija malendi ja malendite listi esitamine FEN-kujul.

instance Show PlPiece
  where

    showsPrec n (Piece pl p)
      = let
          plFun White
            = toUpper
          plFun _
            = id
        in
        (\ (c : cs) -> plFun pl c : cs) . showsPrec 0 p

    showList []
      = ('-' :)
    showList crs
      = foldr1 (.) (map (showsPrec 0) crs)


-- FEN-kujul esitatud mngija malendi vi malendite listi tuvastamine.

instance Read PlPiece
  where

    readsPrec n s@(c : cs)
      | isUpper c
        = [ (Piece White p , t) | (p , t) <- readsPrec n (toLower c : cs)]
      | otherwise
        = [ (Piece Black p , t) | (p , t) <- readsPrec n s]
    readsPrec _ _
      = []

    readList s@(c : cs)
      = let
          readNonEmptyList s@(c : cs)
            = case c of
                ' '
                  -> [([] , s)]
                '-'
                  -> [([] , cs)]
                _
                  -> [
                       (plp : plps , s2)
                     |
                       (plp , s1) <- reads s,
                       (plps , s2) <- readNonEmptyList s1
                     ]
          readNonEmptyList _
            = [([] , "")]
        in
        readNonEmptyList s
    readList _
      = []


{-
Vlja olek (thi vi sisaldab mne mngija malendit).
-}

type State
       = Maybe PlPiece


