-- See moodul defineerib malekigu- ja maleseisutbi ja realiseerib mned 
-- vajalikud nendega seotud funktsioonid.
-- Seisutp on abstraktne selles mttes, et tema konstruktoreid ei ekspordita.

module Position (
         module Square,
         Piece (..),
         allPieces, crownPieces, pieceSyms,
         Player (..),
         op, plDir, relRank,
         PiecesList,
         PlPiece (..),
         State,
         Board (),
         Castling (..),
         clDir, rook,
         Move (..),
         Position (),
         getState, changeState, getPieces, getPlayer, changePlayer,
         getEPSquare, setEPSquare, getCastling, disallowCastling, initPosition,
         showMove, readMove
       ) where


import Square


{-
Moodulid  ListBoard ,  ArrayBoard  ja  FunBoard  realiseerivad malelaua
seisu ja operatsioonid sellel erinevatel viisidel. Kik need moodulid 
ekspordivad abstraktse malelauatbi  Board  ja samade nimedega funktsioonid 
samade operatsioonide sooritamiseks. Neist need, mida see moodul kasutab, on 
jrgmised.
(1) initBoard :: Board 
    annab standardse algseisu.
(2) find :: Int -> Int -> Board -> State 
    leiab antud koordinaatide ja laua seisu phjal viidatud vlja oleku.
(3) change :: State -> Int -> Int -> Board -> Board
    muudab antud seisus antud koordinaatidega viidatud vlja olekut.
(4) findAll :: Player -> Board -> [(Piece , Sq)]
    leiab antud mngija malendite ja asukohtade loetelu antud laua seisus ja 
    vljastab need sorteeritud listina (st kuningas on esimene, siis tulevad
    lipud, siis vankrid jne malenditbi definitsioonis esitatud jrjekorras).
Kommenteerige lahti ks jrgnevast kolmest reast --- see, millega Teie
programm ttab kige paremini.
-}

import ListBoard
-- import ArrayBoard
-- import FunBoard



{-
Vangerdusetp: vangerdus saab olla kas lhike vi pikk.
-}

data Castling
       = Short
       | Long
       deriving (Eq)


-- Vangerduse suunda nitav tisarv.

clDir
  :: Castling -> Int
clDir Short
  = 1
clDir _
  = -1


-- Liin, millel asub vangerdav vanker.

rook
  :: Castling -> Int
rook Short
  = 8
rook _
  = 1


-- Jrgnevad kaks funktsiooni seonduvad ainult FEN genereerimise ja
-- parsimisega.

piece
  :: Castling -> Piece
piece Short
  = King
piece _
  = Queen

clMatch
  :: Player -> Castling -> PlPiece -> Bool
clMatch pl c (Piece pl' p)
  = pl == pl' && piece c == p



{-
Kigu- ja positsioonitp.
-}

data Move
       = Move Piece Sq Sq   -- Malendi nihutamine helt vljalt teisele
                            -- (vib ka la).
       | Crown Sq Sq Piece  -- Etturi lipustumine malendiks
                            -- (vib samuti la).
       | Castling Castling  -- Vangerdus.
       deriving (Eq)


-- Positsioon sisaldab endas jrjekorras jrgmist infot.
-- (1) Malelaua seis.
-- (2) Laual olevate malendite loetelu (saaks ka malelaua seisust ktte, aga 
--     efektiivsuse mttes hoitakse seda eraldi).
-- (3) Kigul olev mngija.
-- (4) Vangerdusiguste loetelu, kus teatavat liiki vangerdamise igust
--     kindlal mngijal vljendab selle mngija vastava tiiva identifitseeriv
--     malend. Tegemist on phimtteliste igustega, mis vljendavad ainult
--     seda, kas potentsiaalselt vangerdavad malendid on juba liikunud vi 
--     ldud vi mitte.
-- (5) Nn "en passant"-vli, kui on. Kui viimane kik oli etturiga
--     algseisust kahe ruudu vrra edasi, siis vli, mida see ettur letas, on
--     antud seisu "en passant"-vli. Kui viimane kik oli midagi muud, siis 
--     antud seisul "en-passant"-vlja ei ole.
-- Seega positsiooni vljad on suuresti samad mis vastaval FEN-kirjel.

data Position
       = Position Board PiecesList Player [PlPiece] (Maybe Sq)
       deriving (Eq)


-- Seisutbi "en-passant"-vlja FEN-kujule teisendamine.

showEP
  :: Maybe Sq -> ShowS
showEP (Just sq)
  = showsPrec 0 sq
showEP _
  = ('-' :)


-- FEN-notatsiooni "en passant"-vlja parsimine seisutbi jaoks.

readEP
  :: ReadS (Maybe Sq)
readEP s@(c : cs)
  = case c of
      '-'
        -> [(Nothing , cs)]
      _
        -> [(Just sq , t) | (sq , t) <- readsPrec 0 s]


-- Malendite loetelu vrskendamine vastavalt konkreetse vlja uuele olekule.

changePieces
  :: State -> Int -> Int -> Board -> PiecesList -> PiecesList
changePieces st i j brd
  = let
      st'
        = find i j brd
    in
    (case st of
      Just (Piece pl p)
        -> add p i j pl
      _
        -> id) .
    (case st' of
      Just (Piece pl p)
        -> remove p i j pl
      _
        -> id)


-- Malendite loetelu leidmine malelaua seisu phjal.

findAllPieces
  :: Board -> PiecesList
findAllPieces
  = let
      fromLists f
        = (f White , f Black)
    in
    fromLists . flip findAll


-- Seisu viimine FEN-kujule.

instance Show Position
  where

    showsPrec n (Position brd _ pl cs ep)
      = showsPrec n brd . (' ' :) .
        showsPrec n pl . (' ' :) .
        showList cs . (' ' :) .
        showEP ep . (" 0 1" ++)


-- FEN-kujul antud seisu parsimine.

instance Read Position
  where

    readsPrec n s
      = [
          (Position brd (findAllPieces brd) pl cr ep , s5)
        |
          (brd , ' ' : s1) <- readsPrec n s,
          (pl , ' ' : s2) <- readsPrec n s1,
          (cr , ' ' : s3) <- readList s2,
          (ep , ' ' : s4) <- readEP s3,
          let s5 = takeWhile (/= ' ') (tail (takeWhile (/= ' ') s4))
        ]


-- Antud vlja oleku leidmine antud seisus.

getState
  :: Sq -> Position -> State
getState sq (Position brd _ _ _ _)
  = find (file sq) (rank sq) brd


-- Antud vlja oleku vrskendamine antud seisus.

changeState
  :: Sq -> State -> Position -> Position
changeState sq st (Position brd ps pl cr ep)
  = let
      x = file sq
      y = rank sq
    in
    Position (change st x y brd) (changePieces st x y brd ps) pl cr ep


-- Antud mngija malendite loetelu leidmine antud seisus.

getPieces
  :: Player -> Position -> [(Piece , Sq)]
getPieces pl (Position _ (wps , bps) _ _ _)
  = case pl of
      White
        -> wps
      _
        -> bps


-- Antud seisus kigul oleva mngija leidmine.

getPlayer
  :: Position -> Player
getPlayer (Position _ _ pl _ _)
  = pl


-- Antud seisus kigul oleva mngija vahetamine.

changePlayer
  :: Position -> Position
changePlayer (Position brd ps pl cr ep)
  = Position brd ps (op pl) cr ep


-- Antud seisu "en-passant"-vlja leidmine.

getEPSquare
  :: Position -> Maybe Sq
getEPSquare (Position _ _ _ _ ep)
  = ep


-- Antud vlja seadmine antud seisu "en-passant"-vljaks.

setEPSquare
  :: Maybe Sq -> Position -> Position
setEPSquare ep (Position brd ps pl cr _)
  = Position brd ps pl cr ep


-- Antud mngija antud liiki vangerdamise iguse kontrollimine antud seisus.

getCastling
  :: Player -> Castling -> Position -> Bool
getCastling pl c (Position _ _ _ cr _)
  = not (null (filter (clMatch pl c) cr))


-- Antud mngija antud liiki vangerdamise iguse thistamine antud seisus.

disallowCastling
  :: Player -> Castling -> Position -> Position
disallowCastling pl c (Position brd ps pl' cr ep)
  = Position brd ps pl' (filter (not . clMatch pl c) cr) ep


-- Standardne algseis.

initPosition
  :: Position
initPosition
  = (\ brd
       -> Position brd (findAllPieces brd) White
            [Piece White King, Piece White Queen, Piece Black King, Piece Black Queen]
              Nothing)
      initBoard


-- Kigu kirjutamine Xboardiga suhtlusel kasutatavas notatsioonis.

showMove
  :: Position -> Move -> String
showMove pos mv
  = case mv of
      Move _ i o
        -> showsPrec 0 i $
           showsPrec 0 o $
           ""
      Crown i o p
        -> showsPrec 0 i $
           showsPrec 0 o $
           showsPrec 0 p $
           ""
      Castling c
        -> let
             rw
               = relRank (getPlayer pos) 1
           in
           showsPrec 0 (square 5 rw) $
           showsPrec 0 (square (5 + 2 * clDir c) rw) $
           ""

-- Xboardiga suhtlusel kasutatavas notatsioonis esitatud kigu hene 
-- parsimine.

readMove
  :: Position -> String -> Move
readMove pos s
  = case s of
      ix : iy : ox : oy : (rest@ ~[cp])
        | null rest
          -> case p of
               King
                 -> case file o - file i of
                      2
                        -> Castling Short
                      -2
                        -> Castling Long
                      _
                        -> Move p i o
               _
                 -> Move p i o
        | otherwise
          -> Crown i o (read rest)
        where
          i = read [ix, iy]
          o = read [ox, oy]
          p = let
                Just (Piece _ p)
                  = getState i pos
              in
              p
      _
        -> error (s ++ ": illegal move!")


