-- See moodul defineerib vimalikud mngu tulemused ja mnguseisu
-- hinnangulised vrtused.

module Result where



import Prelude hiding (negate)


-- Tbiklass, mille esindajatel omab mtet vastandvrtuse leidmise
-- operatsioon.

class (Enum a, Bounded a)
  => Neg a
  where
    negate :: a -> a

    negate x
      = toEnum (fromEnum (maxBound `asTypeOf` x) - fromEnum x)


-- Mngu tulemuse tp.

data Result
       = Lose  -- Kaotus,
       | Draw  -- viik,
       | Win   -- vit.
       deriving (Eq, Bounded, Ord, Show)


-- Mngu tulemuse vljendamine tisarvuga ja mberprdult.
-- Kasutatakse vrdluses heurilistiliste seisuhinnangute poolt antavate 
-- vrtustega.

instance Enum Result
  where

    fromEnum Lose
      = -1
    fromEnum Draw
      = 0
    fromEnum _
      = 1

    toEnum n
      = case compare n 0 of
          LT
            -> Lose
          EQ
            -> Draw
          _
            -> Win


-- Mngu tulemuse vastandtulemuse leidmine.

instance Neg Result
  where

    negate Lose
      = Win
    negate Draw
      = Draw
    negate _
      = Lose


-- Ordering-tbil rakendub vaikimisi-definitsioon.

instance Neg Ordering
  where



-- Seisu vrtuse tp.
-- Seisu vrtus tuleb anda kigul oleva mngija suhtes.

data RealFrac a
  => Value a
       = Estimated a  -- Heuristiline hinnang, vljendatud arvuga tbist  a .
                      -- See arv peaks olema rangelt -1 ja 1 vahel, et oleks 
                      -- kooskla sellega, et kindel kaotus on -1 ja kindel 
                      -- vit 1; seda suurem, mida kasulikum on seis, 
                      -- ihaldatavatel seisudel peaks see olema positiivne ja 
                      -- vlditavatel negatiivne.
       | Definite Result Integer
                      -- Kindel tulemus koos selle vormistamiseks kuluva 
                      -- poolkikude arvuga (st mlema mngija kikude
                      -- arvude summaga).
       deriving (Eq, Show)


-- Seisuvrtuste konverteerimine he dimensiooniga mdetavaks
-- arvuliseks vrtuseks.

toRealFrac
  :: RealFrac a
  => Value a -> a
toRealFrac (Definite res _)      -- Kui leitud seisuvrtus on kindel tulemus, 
  = fromIntegral (fromEnum res)  -- siis vastav hinnanguline vrtus leitakse 
                                 -- vastavalt laldefineeritud vastavusele  fromEnum ,
toRealFrac (Estimated x)         -- vastasel korral
  = x                            -- klbab vana vrtus.


-- Seisuvrtuste vrdlemine.

instance RealFrac a
  => Ord (Value a)
  where

    compare (Definite res m) (Definite ses n)  -- Kahe kindla tulemuse puhul
      = case compare res ses of
          EQ                                   -- (1) kui tulemused on samad,
            -> case res of
                 Lose                          --     siis kaotuse puhul
                   -> compare m n              --     on eelistatavam suurem kikude arv,
                 Win                           --     vidu puhul
                   -> compare n m              --     viksem kikude arv
                 _                             --     ning viigi puhul
                   -> EQ                       --     on kskik,
          c                                    -- (2) kui tulemused on erinevad,
            -> c                               --     vrreldakse tulemusi.
    compare v1               v2  -- lejnud juhtudel 
      = compare (toRealFrac v1) (toRealFrac v2)
                                 -- vrreldakse vastavaid konverteeritud vrtusi.


-- Triviaalne seisuvrtus, milleks on antud mngutulemuse kindel 
-- saavutamine kohe ehk 0 kiguga.
-- Eeldades, et kik heuristilised hinnangud vljenduvad arvudega rangelt 
-- -1 ja 1 vahel, on  immediate Win  eespooltoodud jrjestuses suurim 
-- ning  immediate Lose  vhim vrtus.

immediate
  :: RealFrac a
  => Result -> Value a
immediate res
  = Definite res 0

